// Copyright (c) 2003-2010 Christopher M. Kohlhoff
// Modifications (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#include "icmp/icmpv6packet.h"

#include <iostream>

#include <logfunc.hpp>
#include <boost/scoped_array.hpp>

#include "boost_assert_handler.h"

using namespace std;
using boost::asio::ip::address;
using boost::date_time::time_resolution_traits_adapted64_impl;
using boost::posix_time::ptime;
using boost::posix_time::microsec_clock;
using boost::scoped_array;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// Icmpv6Packet
//-----------------------------------------------------------------------------

/**
 * @brief Default constructor.
 */
Icmpv6Packet::Icmpv6Packet() :
    IpHeader(),
    IcmpPayloadHeader(),
    IcmpPayloadData()
{
}

/**
 * @brief Parameterized constructor.
 *
 * @param icmp_header The ICMP header.
 * @param icmp_data The ICMP payload data.
 */
Icmpv6Packet::Icmpv6Packet(
        const Icmpv6Header &icmp_header,
        const IcmpData &icmp_data
) :
    IpHeader(),
    IcmpPayloadHeader( icmp_header ),
    IcmpPayloadData( icmp_data )
{
}

/**
 * @brief Destructor.
 */
Icmpv6Packet::~Icmpv6Packet()
{
}

/**
 * @brief Obtain the IP header.
 *
 * @return The IP header object.
 */
Ipv6Header Icmpv6Packet::get_ip_header() const
{
    return IpHeader;
}

/**
 * @brief Obtain the ICMP header.
 *
 * @return The ICMP header object.
 */
Icmpv6Header Icmpv6Packet::get_icmp_header() const
{
    return IcmpPayloadHeader;
}

/**
 * @brief Obtain the ICMP payload data.
 *
 * @return The ICMP data.
 */
IcmpData Icmpv6Packet::get_icmp_data() const
{
    return IcmpPayloadData;
}

/**
 * @brief Convenience method to check if this packet, matching the arguments,
 * is a echo reply.
 *
 * @param identifier The identifier.
 * @param sequence_number The sequence number.
 * @param source_address The source address.
 *
 * @return @c true if this packet is a echo reply, or @c false otherwise.
 */
bool Icmpv6Packet::match_echo_reply(
        const uint16_t identifier,
        const uint16_t sequence_number,
        const address &source_address
) const
{
    BOOST_ASSERT( source_address.is_v6() );

    return match( Icmpv6Type_EchoReply, identifier, sequence_number, source_address );
}

/**
 * @brief Convenience method to check if this packet, matching the arguments,
 * is a destination unreachable.
 *
 * @param identifier The identifier.
 * @param sequence_number The sequence number.
 * @param source_address The source address.
 *
 * @return @c true if this packet is a destination unreachable, or @c false
 * otherwise.
 */
bool Icmpv6Packet::match_destination_unreachable(
        const uint16_t identifier,
        const uint16_t sequence_number,
        const address &source_address
) const
{
    BOOST_ASSERT( source_address.is_v6() );

    return match( Icmpv6Type_DestinationUnreachable, identifier, sequence_number, source_address );
}

/**
 * @brief Check if this object matches with all the parameters.
 *
 * @param type The type of ICMP message.
 * @param identifier The identifier.
 * @param sequence_number The sequence number.
 * @param source_address The source address.
 *
 * @return @c true if this matches all the parameters, or @c false if at least
 * one does not match.
 */
bool Icmpv6Packet::match(
        const Icmpv6Type type,
        const uint16_t identifier,
        const uint16_t sequence_number,
        const address &source_address
) const
{
    BOOST_ASSERT( source_address.is_v6() );

    bool type_match = IcmpPayloadHeader.get_type() == type ? true : false;
    bool identifier_match = IcmpPayloadHeader.get_identifier() == identifier ? true: false;
    bool seq_num_match = IcmpPayloadHeader.get_sequence_number() == sequence_number ? true : false;
#ifdef IPV6_DATA_PRESENT_IN_ISTREAM
    // TODO operator>> does not read IpHeader, thus this object is not initialized
    // must check why IPv6 data does not come in the istream like IPv4 data
    bool address_match = IpHeader.get_source_address() == source_address ? true : false;

    return ( type_match && identifier_match && seq_num_match && address_match );
#else

    return ( type_match && identifier_match && seq_num_match );
#endif
}

/**
 * @brief Prints the ICMP echo reply messages.
 *
 * @param bytes_transferred Number of bytes transferred.
 * @param time_packet_sent The time when this packet was sent.
 *
 * @return void
 */
void Icmpv6Packet::print_echo_reply(
        const size_t &bytes_transferred,
        const ptime &time_packet_sent
) const
{
    BOOST_ASSERT( get_icmp_header().get_type() == Icmpv6Type_EchoReply );

    Ipv6Header ipv6_header = get_ip_header();
    Icmpv6Header icmpv6_header = get_icmp_header();

    size_t bytes_received = bytes_transferred - ipv6_header.get_payload_length();
#ifdef IPV6_DATA_PRESENT_IN_ISTREAM
    //TODO WHY IPv6 does not come like IPv4????
    string remote_address = ipv6_header.get_source_address().to_string();
#else
    string remote_address = "?";
#endif

    uint16_t sequence_number = icmpv6_header.get_sequence_number();
    int ttl = ipv6_header.get_hop_limit();
    ptime now = microsec_clock::universal_time();
    time_resolution_traits_adapted64_impl::int_type elapsed_time =
            (now - time_packet_sent).total_milliseconds();

    GlobalLogger.info() << bytes_received << " bytes "
         << "from " << remote_address
         << ": icmp_seq=" << sequence_number
         << " ttl=" << ttl
         << " time=" << elapsed_time << " ms" << endl;
}

/**
 * @brief Prints the destination unreachable messages.
 *
 * @return void
 */
void Icmpv6Packet::print_destination_unreachable() const
{
    BOOST_ASSERT( get_icmp_header().get_type() == Icmpv6Type_DestinationUnreachable );

    Ipv6Header ipv6_hdr = get_ip_header();
    Icmpv6Header icmpv6_hdr = get_icmp_header();

    string local_address = ipv6_hdr.get_destination_address().to_string();
    uint16_t sequence_number = icmpv6_hdr.get_sequence_number();

    GlobalLogger.info() << "From " << local_address
         << " icmp_seq=" << sequence_number
         << " Destination Net Unreachable" << endl;
}

/**
 * @brief Read (part of) the ICMP packet from the input stream @a is
 *
 * @param is The input stream.
 *
 * @return result of the read, currently one of {ok, fail, not enough data}
 */
IcmpPacket::ReadReturnCode Icmpv6Packet::read( istream &is )
{
    if ( !is.good() )
        return IcmpPacket::ReadReturnCode_BAD_STREAM;

#ifdef IPV6_DATA_PRESENT_IN_ISTREAM
    //TODO WHY IPv6 does not come like IPv4????
    // try to read ip header
    is >> IpHeader;

    if ( is.eof() )
        return IcmpPacket::ReadReturnCode_NOT_ENOUGH_DATA;
    else if ( !is.good() )
        return IcmpPacket::ReadReturnCode_BAD_STREAM;
#endif

    // try to read the icmp header
    is >> IcmpPayloadHeader;

    if ( is.eof() )
        return IcmpPacket::ReadReturnCode_NOT_ENOUGH_DATA;
    else if ( !is.good() )
        return IcmpPacket::ReadReturnCode_BAD_STREAM;

    // calculate the size of the ICMP data
    streamsize data_length = static_cast<streamsize>( IpHeader.get_payload_length() )
                           - static_cast<streamsize>( IcmpPayloadHeader.get_header_length() );

    // try to read that amount of data
    if ( data_length < 0 )
    {
        GlobalLogger.error() << "Invalid size for optional ICMP data: "
                             << data_length << endl;
        is.setstate( ios::failbit );
        return IcmpPacket::ReadReturnCode_INVALID_SIZE;
    }
    else if ( data_length > 0 )
    {
        size_t options_size = static_cast<size_t>( data_length );
        scoped_array<uint8_t> scoped_data( new uint8_t[options_size+1] ); // need a 0 after data
        memset(scoped_data.get(), 0, (options_size+1)*sizeof(uint8_t));
        char *char_data = reinterpret_cast<char *>( scoped_data.get() );

        (void) is.read( char_data, data_length );    // leave 0 at the end
        IcmpPayloadData = char_data;  // implicit cast
    }

    if ( is.eof() )
        return IcmpPacket::ReadReturnCode_NOT_ENOUGH_DATA;
    else if ( !is.good() )
        return IcmpPacket::ReadReturnCode_BAD_STREAM;
    else
        return IcmpPacket::ReadReturnCode_OK;
}

/**
 * @brief Write the ICMP packet to the @c ostream.
 *
 * @param os The output stream.
 *
 * @return @c true if the write was successful, or @c false if an error occurred.
 */
bool Icmpv6Packet::write( std::ostream &os ) const
{
    os.clear();

    os << IcmpPayloadHeader << IcmpPayloadData;

    return !os.fail();
}

